/*

Fast image steganography routines and classes

(c)  2009 Alexander Myasnikov

Russia, Vladimir region, Kolchugino

Uses 9 bits of one pixel, 3 bit of one color and 24-bit images.

i.e. for one byte is needed only 0,88 pixels.

WEB: www.darksoftware.narod.ru

E-mail: darksoftware@ya.ru

Free for non-commercial usage

If you like this code and want to use it

in commercial software, please contact author.

 */

#include <stdlib.h>
#include <vcl.h>
#include <stdio.h>
#include "gstego.h"

TIMFile::TIMFile() {

	BitAPos = 0;
	Estimated = 0;
	Injected = false;
	ima_size = 0;
	ima_h = 0;
	ima_w = 0;
	stegotype = 0;
	LengthBitA = 0;
	IMA = (char*)malloc(1);
	BitA = (char*)malloc(1);

}

TIMFile::~TIMFile() {
	free(BitA);
	free(IMA);

}

unsigned char TIMFile::GetBit() {
	if (BitAPos < LengthBitA) {
		unsigned char Result = BitA[BitAPos];
		BitAPos++;
		return Result;
	}
	else {

		int Result = random(256) & 1;
		Injected = true;
		return Result;
	}
}

void TIMFile::PutBit(unsigned char b) {
	if ((LengthBitA - BitAPos) < 32768) {
		LengthBitA += 32768;
		BitA = (unsigned char*)realloc(BitA, LengthBitA);
	}

	BitA[BitAPos] = b;
	BitAPos++;

}

void TIMFile::PutByte(unsigned char b) {

	for (int n = 0; n != 8; n++) {
		BitA[BitAPos] = (b >> n) & 1;
		BitAPos++;
	}

}

void TIMFile::PutInteger(int b) {
	for (int n = 0; n != 32; n++) {
		BitA[BitAPos] = (b >> n) & 1;
		BitAPos++;

	}
}

void TIMFile::PutWord(unsigned short int b) {
	for (int n = 0; n != 16; n++) {
		BitA[BitAPos] = (b >> n) & 1;
		BitAPos++;
	}
}

unsigned char TIMFile::GetByte() {
	unsigned char Result = 0;

	BitAPos += 8;

	for (int n = 1; n != 9; n++) {
		Result = (Result << 1) + BitA[BitAPos - n];
	}

	return Result;

}

int TIMFile::GetInteger() {
	int Result = 0;

	BitAPos += 32;

	for (int n = 1; n != 33; n++) {
		Result = (Result << 1) + BitA[BitAPos - n];
	}

	return Result;
}

unsigned short int TIMFile::GetWord() {
	unsigned short int Result = 0;

	BitAPos += 16;

	for (int n = 1; n != 17; n++) {
		Result = (Result << 1) + BitA[BitAPos - n];
	}

	return Result;

}

void TIMFile::LoadToBA(String fn) {

	unsigned char d;

	TFileStream *FS = new TFileStream(fn, fmOpenRead);
	LengthBitA = FS->Size * 8 + 48;
	BitA = (unsigned char*)realloc(BitA, LengthBitA);
	BitAPos = 0;
	PutWord(Header);
	PutInteger(FS->Size);

	for (int i = 0; i != FS->Size; i++) {
		FS->Read(&d, 1);
		PutByte(d);
	}

	delete FS;
}

void TIMFile::SaveToBA(String fn) {

	unsigned char d;

	TFileStream *FS = new TFileStream(fn, fmCreate);
	BitAPos = 16;
	int size = GetInteger();

	for (int i = 0; i != size; i++) {
		d = GetByte();
		FS->Write(&d, 1);
	}

	delete FS;
}

bool TIMFile::CheckHeader(String fn) {
	LoadImage(fn);
	Catch();
	BitAPos = 0;
	if (LengthBitA > 16) {
		return(GetWord() == Header);
	}
	else {
		return false;
	}
}

void TIMFile::LoadImage(String fn) {

	unsigned char *_ima;

	TMemoryStream *bmpstream;

	Graphics::TBitmap *bmp = new Graphics::TBitmap;
	bmp->PixelFormat = pf24bit;

	bmp->LoadFromFile(fn);

	ima_w = bmp->Width;
	ima_h = bmp->Height;

	ima_size = ima_w * ima_h * 3;

	IMA = (unsigned char*)realloc(IMA, ima_size);

	_ima = IMA;

	for (int i = 0; i != ima_h; i++) {

		memcpy(IMA, bmp->ScanLine[i], ima_w * 3);
		IMA += ima_w * 3;

	}

	IMA = _ima;

	bmp->Dormant();
	delete bmp;

	switch(stegotype) {
	case 0:
		Estimated = ((ima_w * ima_h * 12) / 8) - 48;
		break;
	case 1:
		Estimated = ((ima_w * ima_h * 9) / 8) - 48;
		break;
	case 2:
		Estimated = ((ima_w * ima_h * 6) / 8) - 48;
		break;
	case 3:
		Estimated = ((ima_w * ima_h * 3) / 8) - 48;
		break;

	}

}

void TIMFile::SaveImage(String fn) {

	unsigned char * _ima;
	TMemoryStream *BmpStream;

	Graphics::TBitmap *bmp = new Graphics::TBitmap();
	bmp->PixelFormat = pf24bit;
	bmp->Width = ima_w;
	bmp->Height = ima_h;
	_ima = IMA;

	for (int i = 0; i != ima_h; i++) {
		memcpy(bmp->ScanLine[i], IMA, ima_w * 3);
		IMA += ima_w * 3;

	}

	IMA = _ima;

	bmp->SaveToFile(fn);

	bmp->Dormant();
	delete bmp;

}

int TIMFile::changebits(unsigned char v) {
	unsigned char x[8];

	unsigned char Result = 0;

	for (int i = 0; i != 8; i++) {
		x[i] = (v >> i) & 1;
	}

	switch(stegotype) {
	case 0:

		x[0] = GetBit();
		x[1] = GetBit();
		x[2] = GetBit();
		x[3] = GetBit();
		break;

	case 1:

		x[0] = GetBit();
		x[1] = GetBit();
		x[2] = GetBit();
		break;
	case 2:

		x[0] = GetBit();
		x[1] = GetBit();
		break;
	case 3:

		x[0] = GetBit();
		break;

	}

	for (int i = 7; i != -1; i--) {
		Result = (Result << 1) + x[i];
	}
	return Result;
}

void TIMFile::Inject() {
	int i;
	unsigned char *_ima;

	{

		BitAPos = 0;

		_ima = IMA;

		for (int i = 0; i != ima_size; i++) {

			*IMA = changebits(*IMA);

			IMA++;

		}

		IMA = _ima;

	}

}

void TIMFile::getbits(unsigned char v) {
	unsigned char x[8];

	for (int i = 0; i != 8; i++) {
		x[i] = (v >> i) & 1;
	}

	switch(stegotype) {
	case 0:
		PutBit(x[0]);
		PutBit(x[1]);
		PutBit(x[2]);
		PutBit(x[3]);
		break;

	case 1:

		PutBit(x[0]);
		PutBit(x[1]);
		PutBit(x[2]);
		break;

	case 2:

		PutBit(x[0]);
		PutBit(x[1]);
		break;

	case 3:

		PutBit(x[0]);

		break;
	}

}

void TIMFile::Catch() {
	unsigned char *_ima;

	BitAPos = 0;

	_ima = IMA;

	for (int i = 0; i != ima_size; i++) {

		getbits(*IMA);

		IMA += 1;

	}

	IMA = _ima;

}
